
/**
Testprogramm fpr ATmega328P, geschrieben in avr-gcc.

Gemessen wird die Periodendauer eines TTL-Signals mit einer Frequenz von
ca. 1 - 2 Hz, das an Pin ICP1 = Pin 14 in einen ATmega238P eingespeist
wird, mittels Timer 1, der im 16 Bit Input Capture Modus betrieben und
per Software auf 32 Bit "verlängert" wird.

Ausgegeben wird die Zeit in Ticks. Für die Dauer eines Ticks gilt
t_tick = 1 / f_quarz, d. h. bei 16 MHz Quarzfrequenz dauert ein Tick 62.5 ns.

Zusätzlich wird an Pin OC0A = Pin 12 ein TTL-Signal mit einer Frequenz
von einem Zehntel der Quarzfrequenz ausgegeben, so dass die Quarzfrequenz
mit einem Frequenzzähler ermittelt werden kann.

Mit einem Terminalprogramm (z. B. "minicom") kann ähnliches empfangen werden:

...
  1669204654   8652439      5728
  1677850933   8646279     -6160
  1686493968   8643035     -3244
  1695138053   8644085      1050
  1703783894   8645841      1756
  1712430843   8646949      1108
  1721075495   8644652     -2297
  1729722220   8646725      2073
  1738377495   8655275      8550
  1747031046   8653551     -1724
  1755684181   8653135      -416
...

Spalte 1

	Absolute Zeit in Ticks. Überlauf alle 4.5 min.

Spalte 2

	Zeit in Ticks zwischen zwei steigenden / positiven Flanken an Pin ICP1

Spalte 3

	Unterschied in Ticks zwischen zwei der obigen Meßwerte

Das Programm scheint zu funktionieren - dafür, dass es das aber in jedem Betriebs-
zustand wirklich zu 100 % tut, würde ich meine Hand aber nicht ins Feuer legen. Zur
Variablen "Problem" siehe meine o. g. Überlegungen.
**/

#include <stdio.h>
#include <stdint.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#define BAUD 38400UL

#define MYUBRR (((F_CPU + BAUD * 8LU) / (BAUD * 16LU) - 1UL))

static FILE mystdout = { 0 };

volatile uint16_t t1_overflow = 0;

volatile uint16_t temp_t1_overflow;

volatile uint16_t temp_icr1;

volatile uint8_t problem;

volatile uint8_t flag = 0;

/********************************************************************************/

static int uart_putchar (char c, FILE *stream)
	{
	loop_until_bit_is_set (UCSR0A, UDRE0);

	UDR0 = c;

	return 0;
	}

/********************************************************************************/

void uart_init (void)
	{
	DDRD |= (uint8_t) 0b10000000;		/// TXD = Ausgang

	UBRR0H = (uint8_t) (MYUBRR >> 8);	/// Baudrate

	UBRR0L = (uint8_t) (MYUBRR & 0xFF);

	UCSR0B = (1<<RXEN0) | (1<<TXEN0);	/// RX und TX Enable

	UCSR0C = (3 << UCSZ00);			/// 8n1

	/** Informationsausgabe mittels "printf" ermöglichen **/

	fdev_setup_stream (&mystdout, uart_putchar, NULL, _FDEV_SETUP_WRITE);

	stdout = &mystdout;
	}

/********************************************************************************/

void t0_ctc (uint8_t x)
	{
	DDRD |= (1 << PD6);		/// PB6 = Ausgang

	TCCR0A |= (1 << WGM01);		/// Clear Timer on Compare Match (CTC) Mode
	TCCR0A |= (1 << COM0A0);	/// Toggle PB6 on compare match
	TCCR0B |= (1 << CS00);		/// Prescaler: div 1

	OCR0A = x;			/// Frequenzteiler

	TCNT0 = 0;			/// Counter Reset + Start
	}

/********************************************************************************/

ISR (TIMER1_OVF_vect)
	{
	uint8_t sreg = SREG;

	cli ();

	t1_overflow++;

	SREG = sreg;
	}

/********************************************************************************/

ISR (TIMER1_CAPT_vect)
	{
	uint8_t sreg = SREG;

	cli ();

	if ((TIFR1 & (1<<TOV1)) && (ICR1 < 0x8000))
			problem = 1;
		else	problem = 0;

	temp_t1_overflow = t1_overflow;

	temp_icr1 = ICR1;

	flag = 1;

	SREG = sreg;
	}

/********************************************************************************/

void t1_init (void)
	{
	/** Input capture Pin ICP1 (Pin 14) = Eingang **/

	DDRB &= 0b11111110;

	/** PRR – Power Reduction Register (S. 54) **/

	PRR &= 0b11110111;

	/** TCCR1A – Timer/Counter1 Control Register A (S. 140) **/

	TCCR1A = 0;

	/** TCCR1B – Timer/Counter1 Control Register B (S. 142) **/

	TCCR1B = 0b11000001;

	/** TIMSK1 – Timer/Counter1 Interrupt Mask Register (S. 144) **/

	TIMSK1 = 0b00100001;

	/*** TIFR1 – Timer/Counter1 Interrupt Flag Register (S. 145) **/

	TIFR1 &= 0b00100111;

	/** Interrupts global ein **/

	sei ();

	}

/********************************************************************************/

void capture (void)
	{
	uint32_t neuer_wert, alter_wert = 0;

	int32_t neue_differenz, alte_differenz = 0, diff_diff;

	uart_init ();

	t0_ctc (10 / 2 - 1); /// f_clk / 10 für Frequenzzähler

	t1_init ();

	while (1)
		{
		while (flag == 0)
			;

		/** Absolute Zeit. Überlauf alle 2^32 / 16 MHz = ca. 4.5 min. **/

		neuer_wert = ((uint32_t) temp_icr1) + ((uint32_t) temp_t1_overflow << 16);

		if (problem == 1)
				{
				neuer_wert += (uint32_t) 0x10000;

				printf ("* ");
				}
			else	{
				printf ("  ");
				}

		printf ("%10lu", neuer_wert);

		/** Periodendauer zwischen zwei steigenden / positiven Flanken an Pin ICP1 **/

		neue_differenz = (int32_t) neuer_wert - (int32_t) alter_wert;

		printf ("%10li", neue_differenz);

		/** Differenz der obigen Meßwerte **/

		diff_diff = neue_differenz - alte_differenz;

		printf ("%10li\r\n", diff_diff);

		/** Alte Werte merken **/

		alter_wert = neuer_wert;

		alte_differenz = neue_differenz;

		/** Ereignis-Flag zurücksetzen **/

		flag = 0;
		}
	}

/********************************************************************************/

void info (void)
	{
	unsigned char i = 0;

	uart_init ();

	while (1)
		{
		printf ("%d\r\n", i++);

		_delay_ms (1000);
		}
	}

/********************************************************************************/

int main (void)
	{
	capture ();

	return 0;
	}
